from mpm.model import *
import shutil
import pathlib
import logging
import hashlib
import os
import semantic_version

logger = logging.getLogger('mpm.server')

class Server:
    def __init__(self, name, config):
        self.name = name
        self.config = config
        self.path = config['path']
        self.pluginPath = self.path+'/plugins'

    def exists(self):
        return os.path.exists(self.path) and os.path.exists(self.pluginPath)

    def plugins(self):
        return [PluginSpec(p['name'], p.get('version', '*')) for p in self.config['plugins']]

    def add_plugin(self, pluginSpec):
        for plugin in self.config['plugins']:
            if plugin['name'] == pluginSpec.name:
                raise KeyError("Cannot add plugin multiple times.")
        self.config['plugins'].append({'name': pluginSpec.name, 'version': str(pluginSpec.versionSpec)})

    def set_plugin_version(self, pluginName, versionSpec):
        for plugin in self.config['plugins']:
            if plugin['name'] == pluginName:
                plugin['version'] = str(versionSpec)

    def pluginStates(self, repos):
        managedPluginFilenames = []
        for plugin in self.plugins():
            logger.debug("Checking %r", plugin)
            compatibleVersions = []

            currentVersion = self.currentVersionForPlugin(plugin.name, repos)
            logger.debug('Current version: %s', currentVersion)
            if currentVersion is not None:
                if plugin.versionSpec.match(currentVersion.version):
                    compatibleVersions.append(currentVersion)
                managedPluginFilenames.append(os.path.basename(currentVersion.path))

            for repo in repos:
                for repoPlugin in repo.plugins():
                    if repoPlugin.name == plugin.name:
                        logger.debug("%r -> %r", repoPlugin.version, plugin.versionSpec)
                        logger.debug(plugin.versionSpec.match(repoPlugin.version))
                        if plugin.versionSpec.match(repoPlugin.version):
                            compatibleVersions.append(repoPlugin)
            if len(compatibleVersions) == 0:
                logger.debug('No compatible versions found for %s', plugin)
                yield MissingVersions(plugin, currentVersion)
            else:
                preferredVersion = list(reversed(sorted(compatibleVersions)))[0]
                logger.debug('Wanted %r, found %r', plugin, preferredVersion)
                if currentVersion is not None and preferredVersion == currentVersion:
                    logger.debug("Already installed: %r", currentVersion)
                    yield Installed(plugin, currentVersion)
                else:
                    logger.debug("Update available: %r -> %r", plugin, preferredVersion)
                    yield Outdated(plugin, currentVersion, preferredVersion)

        otherPlugins = os.listdir(self.pluginPath)
        for pluginFile in otherPlugins:
            if os.path.isfile(os.path.join(self.pluginPath, pluginFile)) and pluginFile not in managedPluginFilenames:
                logger.debug("Unmanaged file: %s", pluginFile)
                yield UnmanagedFile(pluginFile)

    def currentVersionForPlugin(self, pluginName, repos=[]):
        plugins = os.listdir(self.pluginPath)
        for pluginJar in plugins:
            if pluginJar.startswith(pluginName) and pluginJar.endswith(".jar"):
                if pluginJar == pluginName + ".jar":
                    logger.debug("Running checksum search to find version of %s", pluginJar)
                    pluginPath = os.path.join(self.pluginPath, pluginJar)
                    with open(pluginPath, 'rb') as fd:
                        installedHash = hashlib.file_digest(fd, 'sha256')
                    pluginName = pluginJar.split('.jar')[0]
                    defaultVersion = semantic_version.Version("0.0.0")
                    for repo in repos:
                        for plugin in repo.plugins():
                            if plugin.name == pluginName:
                                with open(plugin.path, 'rb') as fd:
                                    foundHash = hashlib.file_digest(fd, 'sha256')
                                if foundHash.hexdigest() == installedHash.hexdigest():
                                    return Plugin(pluginPath,
                                                  forceName=pluginName,
                                                  forceVersion = plugin.version)
                    return Plugin(pluginPath,
                                  forceName=pluginName,
                                  forceVersion=defaultVersion)
                else:
                    try:
                        return Plugin(os.path.join(self.pluginPath, pluginJar))
                    except ValueError:
                        continue

    def syncToVersion(self, plugin, repos=[]):
        dest = os.path.join(self.pluginPath, plugin.normalizedFilename())
        logger.debug("Syncing %s to %s", dest, plugin)
        current = self.currentVersionForPlugin(plugin.name, repos)
        if current is not None:
            logger.debug("Removing current version %r", current)
            current.unlink()

        try:
            os.link(plugin.path, dest)
            logger.debug("Plugin synced")
        except Exception:
            logger.debug("Hardlink not possible. Copying instead.")
            shutil.copyfile(plugin.path, dest)
